"""
Summary
-------

A simple python API and package that can be copy / pasted / replaced.

_This description was sourced from the docstring of the root \
level `__init__.py` file of the `ft3.template` subpackage._

"""

__all__ = (
    'api',
    'pkg'
    )

from . import api
from . import pkg
