"""
Summary
-------

A simple python package that can be copy / pasted / replaced.

"""

__all__ = (
    'clients',
    'enm',
    'exc',
    'lib',
    'obj',
    'typ'
    )

from . import clients
from . import enm
from . import exc
from . import lib
from . import obj
from . import typ
